"""Widget"""

from typing import DefaultDict, Dict, Optional, Sequence, Union

import pandas as pd

from .reg import Reg
from .utils import astype
from .vocab import Vocab


def dropdown(
    nickname: str,
    options: Union[Sequence, pd.Series],
    include: str = "",
    exclude: str = "",
):
    """Makes widget that is used by class.

    Options are filtered by a regular expression written to reflect the include and exclude inputs, and these are updated when changed and shown in the dropdown. The user should select using `command` or `control` to make multiple options. Then push the "save" button when the nickname and selected options from the dropdown menu are the variables you want to include exactly in a future regular expression search.

    Parameters
    ----------
    nickname: str
        nickname to associate with the Vocab class vocabulary entry from this, e.g., "temp". Inputting this to the function creates a text box for the user to enter it into.
    options: Sequence
        strings to select from in the dropdown widget. Will be filtered by include and exclude inputs.
    include: str
        include must be in options values for them to show in the dropdown. Will update as more are input. To input more than one, join separate strings with "|". For example, to search on both "temperature" and "sea_water", input "temperature|sea_water".
    exclude: str
        exclude must not be in options values for them to show in the dropdown. Will update as more are input. To input more than one, join separate strings with "|". For example, to exclude both "temperature" and "sea_water", input "temperature|sea_water".
    """
    import ipywidgets as widgets

    reg = Reg(include=include, exclude=exclude)
    print("Regular expression: ", reg.pattern())
    options = astype(options, pd.Series)
    options2 = options[options.str.match(reg.pattern())]

    widg = widgets.SelectMultiple(
        options=options2,
        value=[] if len(options2) == 0 else [options2.iloc[0]],
        rows=10,
        description="Options",
        disabled=False,
        layout=widgets.Layout(width="50%"),
    )
    return widg


class Selector(object):
    """Coordinates interaction with dropdown widget to make simple vocabularies.

    Options are filtered by a regular expression written to reflect the include and exclude inputs, and these are updated when changed and shown in the dropdown. The user should select using `command` or `control` to make multiple options. Then push the "save" button when the nickname and selected options from the dropdown menu are the variables you want to include exactly in a future regular expression search.

    Examples
    --------

    Show widget with a short list of options. Input a nickname and press button to save an entry to the running vocabulary in the object:

    >>> import cf_pandas as cpf
    >>> sel = cfp.Selector(options=["var1", "var2", "var3"])
    >>> sel

    See resulting vocabulary with:

    >>> sel.vocab
    """

    def __init__(
        self,
        options: Sequence,
        vocab: Optional[Vocab] = None,
        nickname_in: str = "",
        include_in: str = "",
        exclude_in: str = "",
    ):
        """Initialize Selector object.

        Parameters
        ----------
        options: Sequence
            strings to select from in the dropdown widget. Will be filtered by include and exclude  inputs.
        vocab: Vocab object
            Defaults to None. A vocabulary will be created as part of using this widget. However, instead a vocabulary can be input via this argument and then will be amended with the entries made with the widget.
        nickname_in: str
            Default nickname, used for initial value, useful for testing
        include_in: str
            Default include, used for initial value, useful for testing
        exclude_in: str
            Default exclude, used for initial value, useful for testing
        """
        import ipywidgets as widgets

        # create an output widget in order to show output instead of going to log
        self.output = widgets.Output()

        if vocab is None:
            self.vocab = Vocab()
        else:
            self.vocab = vocab

        self.dropdown_values: Sequence = []
        self.include = include_in
        self.exclude = exclude_in

        self.button_save = widgets.Button(description="Press to save")

        self.nickname_text = widgets.Text(value=nickname_in)
        self.nickname = self.nickname_text.value

        self.dropdown = widgets.interact(
            dropdown,
            options=widgets.fixed(options),
            nickname=self.nickname,
            include=self.include,
            exclude=self.exclude,
        )
        from IPython.display import display

        display(self.button_save)

        self.button_save.on_click(self.button_pressed)
        display(self.output)

    def button_pressed(self, *args):
        """Saves a new entry in the catalog when button is pressed."""

        # print vocab
        # clear the output on every click of randomize self.val
        self.output.clear_output()

        # execute function so it gets captured in output widget view
        with self.output:
            if self.dropdown.widget.kwargs["nickname"] == "":
                raise KeyError("Must input nickname to make entry.")

            # regular expressions to put into entries: exact matching
            res = [
                Reg(include_exact=exp).pattern()
                for exp in self.dropdown.widget.result.value
            ]
            self.vocab.make_entry(
                self.dropdown.widget.kwargs["nickname"], res, attr="standard_name"
            )
            print("Vocabulary: ", self.vocab)

    # def return_list(self):
    #     """Return list of variables selected."""

    #     self._variables.append(self.dropdown.widget.result.value)



"""Test cf-pandas."""

from unittest import mock

import numpy as np
import pandas as pd
import pytest
import requests

import cf_pandas as cfp

criteria = {
    "wind_s": {
        "standard_name": "wind_speed$",
    },
    "temp2": {
        "standard_name": "temp$",
    },
    "salt2": {"standard_name": "sal$"},
}


def test_options():
    # test for inputting a nonexistent option
    with pytest.raises(ValueError):
        cfp.set_options(DISPLAY_WIDTH=80)


def test_validate():
    df = pd.DataFrame(
        columns=[
            "temp",
        ]
    )
    with pytest.raises(AttributeError):
        df.cf.keys()


def test_match_criteria_key_accessor():

    df = pd.DataFrame(
        columns=[
            "temp",
            "wind_speed",
            "wind_speed (m/s)",
            "WIND_SPEED",
            "wind_speed_status",
            "longitude (degrees_east)",
            "X",
            "latitude",
            "time",
        ]
    )

    # test accessor with set_options criteria
    with cfp.set_options(custom_criteria=criteria):
        assert sorted(df.cf["wind_s"].columns) == ["wind_speed", "wind_speed (m/s)"]
        assert isinstance(df.cf["wind_s"], pd.DataFrame)
        assert df.cf["temp2"].name == "temp"
        assert isinstance(df.cf["temp2"], pd.Series)
        assert df.cf["longitude"].name == "longitude (degrees_east)"
        assert df.cf.custom_keys["temp2"] == ["temp"]
        assert sorted(df.cf.custom_keys["wind_s"]) == ["wind_speed", "wind_speed (m/s)"]
        assert sorted(df.cf.axes) == ["T", "X"]
        assert sorted(df.cf.coordinates) == ["latitude", "longitude", "time"]
        assert sorted(df.cf.keys()) == [
            "T",
            "X",
            "latitude",
            "longitude",
            "temp2",
            "time",
            "wind_s",
        ]
        assert "X" in df.cf
        assert "Y" not in df.cf
        assert "salt2" not in df.cf


@mock.patch("requests.get")
def test_standard_names(mock_requests):

    resp = requests.models.Response
    resp.content = b"""<?xml version="1.0"?>\n<standard_name_table xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="cf-standard-name-table-1.1.xsd">\n   <version_number>79</version_number>\n   <last_modified>2022-03-19T15:25:54Z</last_modified>\n   <institution>Centre for Environmental Data Analysis</institution>\n   <contact>support@ceda.ac.uk</contact>\n\n  \n   <entry id="longitude">\n      </entry>\n  \n   <entry id="wind_speed">\n  """
    mock_requests.return_value = resp
    df = pd.DataFrame(
        columns=[
            "temp",
            "wind_speed",
            "wind_speed (m/s)",
            "WIND_SPEED",
            "wind_speed_status",
            "longitude (degrees_east)",
            "X",
            "latitude",
            "time",
        ]
    )
    assert df.cf.standard_names["longitude"] == ["longitude (degrees_east)"]
    assert sorted(df.cf.standard_names["wind_speed"]) == [
        "wind_speed",
        "wind_speed (m/s)",
    ]


def test_set_item():
    df = pd.DataFrame(
        columns=[
            "temp",
            "wind_speed",
            "wind_speed (m/s)",
            "WIND_SPEED",
            "wind_speed_status",
            "longitude",
            "latitude",
            "time",
        ]
    )
    with cfp.set_options(custom_criteria=criteria):
        df.cf["temp"] = np.arange(8)
        assert all(df.cf["temp"].values == np.arange(8))
        df.cf["longitude"] = np.arange(8)
        assert all(df.cf["longitude"].values == np.arange(8))


def test_get_by_guess_regex():
    df = pd.DataFrame(columns=["lon", "lat", "min"])
    assert df.cf["longitude"].name == "lon"
    assert df.cf["latitude"].name == "lat"
    assert df.cf["time"].name == "min"



"""Test cf-pandas utils."""

from unittest import mock

import pandas as pd
import requests

import cf_pandas as cfp

criteria = {
    "wind_s": {
        "standard_name": "wind_speed$",
    },
}


def test_match_criteria_key():

    vals = ["wind_speed", "WIND_SPEED", "wind_speed_status"]

    # test function with set_options criteria
    with cfp.set_options(custom_criteria=criteria):
        assert cfp.match_criteria_key(vals, ["wind_s"]) == ["wind_speed"]

    # test function with input criteria
    assert cfp.match_criteria_key(vals, ["wind_s"], criteria) == ["wind_speed"]


def test_match_criteria_key_split():

    vals = ["wind_speed (m/s)", "WIND_SPEED", "wind_speed_status"]

    # test function with set_options criteria
    with cfp.set_options(custom_criteria=criteria):
        assert cfp.match_criteria_key(vals, ["wind_s"], split=True) == [
            "wind_speed (m/s)"
        ]

    # test function with input criteria
    assert cfp.match_criteria_key(vals, ["wind_s"], criteria, split=True) == [
        "wind_speed (m/s)"
    ]


@mock.patch("requests.get")
def test_standard_names(mock_requests):

    resp = requests.models.Response
    resp.content = b"""<?xml version="1.0"?>\n<standard_name_table xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="cf-standard-name-table-1.1.xsd">\n   <version_number>79</version_number>\n   <last_modified>2022-03-19T15:25:54Z</last_modified>\n   <institution>Centre for Environmental Data Analysis</institution>\n   <contact>support@ceda.ac.uk</contact>\n\n  \n   <entry id="longitude">\n      </entry>\n  \n   <entry id="wind_speed">\n  """
    mock_requests.return_value = resp
    names = cfp.standard_names()
    assert "wind_speed" in names


def test__is_datetime_like():
    df = pd.DataFrame()
    df["time"] = pd.date_range(start="2001-1-1", end="2001-1-5", freq="1D")
    assert cfp.utils._is_datetime_like(df["time"])

    df = pd.DataFrame()
    df["time"] = ["2001-1-1", "2001-1-2", "2001-1-3"]
    assert not cfp.utils._is_datetime_like(df["time"])
    assert cfp.utils._is_datetime_like(pd.to_datetime(df["time"]))



"""
cf-pandas: an accessor for pandas objects that interprets CF attributes
"""

from pkg_resources import DistributionNotFound, get_distribution

from .accessor import CFAccessor  # noqa
from .options import set_options  # noqa
from .reg import Reg
from .utils import always_iterable, astype, match_criteria_key, standard_names
from .vocab import Vocab, merge
from .widget import Selector, dropdown

try:
    __version__ = get_distribution("cf-pandas").version
except DistributionNotFound:
    # package is not installed
    __version__ = "unknown"



"""
From cf-xarray.
"""

import itertools
from collections import ChainMap
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import pandas as pd
from pandas import DataFrame, Series

import cf_pandas as cfp

from .criteria import coordinate_criteria, guess_regex
from .options import OPTIONS
from .utils import (
    _is_datetime_like,
    always_iterable,
    match_criteria_key,
    set_up_criteria,
)
from .vocab import Vocab

#:  `axis` names understood by cf_xarray
_AXIS_NAMES = ("X", "Y", "Z", "T")

#:  `coordinate` types understood by cf_xarray.
_COORD_NAMES = ("longitude", "latitude", "vertical", "time")

# Type for Mapper functions
Mapper = Callable[[DataFrame, str], List[str]]


try:
    # delete the accessor to avoid warning
    del pd.DataFrame.cf
except AttributeError:
    pass


@pd.api.extensions.register_dataframe_accessor("cf")
class CFAccessor:
    """Dataframe accessor analogous to cf-xarray accessor."""

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        """what is necessary for basic use."""

        # verify that necessary keys are present. Z would also be nice but might be missing.
        # but don't use the accessor to check
        keys = ["T", "longitude", "latitude"]
        missing_keys = [key for key in keys if len(_get_axis_coord(obj, key)) == 0]
        if len(missing_keys) > 0:
            raise AttributeError(
                f'{"longitude", "latitude", "time"} must be identifiable in DataFrame but {missing_keys} are missing.'
            )

        # for key in keys:
        #     if len(_get_axis_coord(obj, "T")) == 0:

        # if (len(_get_axis_coord(obj, "T")) == 0) or (len(_get_axis_coord(obj, "longitude")) == 0)

        # if not {"longitude", "latitude", "time"} <= obj.cf.coordinates():
        #     raise AttributeError(f'{"longitude", "latitude", "time"} must be identifiable in DataFrame but recognized keys are {obj.cf.keys()}.')

    def __getitem__(self, key: str) -> Union[pd.Series, pd.DataFrame]:
        """Select columns or columns by alias.

        If one column matches key, return a Series. Otherwise return a DataFrame.

        Parameters
        ----------
        key: str
            key in custom criteria/vocabulary to match with columns of DataFrame, or in axes or coordinates.

        Returns
        -------
        Series, DataFrame
            with matching column(s) included.

        Example
        -------
        >>> df.cf[alias]
        """

        # if key is a coordinate or axes, use a different method to match
        valid_keys = _COORD_NAMES + _AXIS_NAMES
        if key in valid_keys:
            col_names = _get_axis_coord(self._obj, key)

        else:
            col_names = _get_custom_criteria(self._obj, key)

        # return series
        if len(col_names) == 1:
            return self._obj[col_names[0]]
        # return DataFrame
        elif len(col_names) > 1:
            return self._obj[col_names]
        else:
            raise ValueError("Some error has occurred.")

    def __setitem__(self, key: str, values: Union[Sequence, Series]):
        """Set column by alias.

        Parameters
        ----------
        key: str
            key in custom criteria/vocabulary to match with columns of DataFrame, or in axes or coordinates.
        values : Union[Sequence, pd.Series]
            Values to set into object.

        Raises
        ------
        ValueError
            Can only set one column at once.
        """

        col = self.__getitem__(key)
        if isinstance(col, Series):
            self._obj[col.name] = values
            # return self._obj[col.name]
        elif col is None:
            # make new column
            self._obj[key] = values
            # return self._obj[key]
        else:
            raise ValueError("Setting item only works if key matches one column only.")

    def __contains__(self, item: str) -> bool:
        """
        Check whether item is a valid key for indexing with .cf
        """
        return item in self.keys()

    def keys(self) -> Set[str]:
        """
        Utility function that returns valid keys for .cf[].

        This is useful for checking whether a key is valid for indexing, i.e.
        that the attributes necessary to allow indexing by that key exist.

        Returns
        -------
        set
            Set of valid key names that can be used with __getitem__ or .cf[key].
        """

        varnames = list(self.axes) + list(self.coordinates)
        try:
            # see which custom keys have matched values in object
            matched_keys = [
                key for key, val in self.custom_keys.items() if len(val) > 0
            ]
            varnames.extend(matched_keys)
        except ValueError:
            # don't have criteria defined, then no custom keys to report
            pass
        # varnames.extend(list(self.cell_measures))
        # varnames.extend(list(self.standard_names))
        # varnames.extend(list(self.cf_roles))

        return set(varnames)

    @property
    def axes(self) -> Dict[str, List[str]]:
        """
        Property that returns a dictionary mapping valid Axis standard names for ``.cf[]``
        to variable names.

        This is useful for checking whether a key is valid for indexing, i.e.
        that the attributes necessary to allow indexing by that key exist.
        It will return the Axis names ``("X", "Y", "Z", "T")``
        present in ``.columns``.

        Returns
        -------
        dict
            Dictionary with keys that can be used with ``__getitem__`` or as ``.cf[key]``.
            Keys will be the appropriate subset of ("X", "Y", "Z", "T").
            Values are lists of variable names that match that particular key.
        """
        # vardict = {key: self.__getitem__(key) for key in _AXIS_NAMES}
        vardict = {key: _get_all(self._obj, key) for key in _AXIS_NAMES}
        return {k: sorted(v) for k, v in vardict.items() if v}

    @property
    def coordinates(self) -> Dict[str, List[str]]:
        """
        Property that returns a dictionary mapping valid Coordinate standard names for ``.cf[]``
        to variable names.

        This is useful for checking whether a key is valid for indexing, i.e.
        that the attributes necessary to allow indexing by that key exist.
        It will return the Coordinate names ``("latitude", "longitude", "vertical", "time")``
        present in ``.columns``.

        Returns
        -------
        dict
            Dictionary of valid Coordinate names that can be used with ``__getitem__`` or ``.cf[key]``.
            Keys will be the appropriate subset of ``("latitude", "longitude", "vertical", "time")``.
            Values are lists of variable names that match that particular key.
        """
        # vardict = {key: self.__getitem__(key) for key in _COORD_NAMES}
        vardict = {key: _get_all(self._obj, key) for key in _COORD_NAMES}

        return {k: sorted(v) for k, v in vardict.items() if v}

    @property
    def custom_keys(self):
        """
        Returns a dictionary mapping criteria keys to variable names.

        Returns
        -------
        dict
            Dictionary mapping criteria keys to variable names.

        Notes
        -----
        Need to use this with context manager version of providing custom_criteria.
        """

        custom_criteria = set_up_criteria()
        vardict = {
            key: _get_custom_criteria(self._obj, key) for key in custom_criteria.keys()
        }

        return vardict

    @property
    def standard_names(self):
        """
        Returns a dictionary mapping standard_names to variable names, if there is a match. Compares with all cf-standard names.

        Returns
        -------
        dict
            Dictionary mapping standard_names to variable names.

        Notes
        -----
        This is not the same as the cf-xarray accessor method of the same name, which searches for variables with standard_name attributes and surfaces those values to map to the variable name.
        """

        names = cfp.standard_names()

        vardict = {}
        for key in names:
            local_criteria = Vocab().make_entry(key, f"{key}$")
            key_match = _get_custom_criteria(
                self._obj, key, criteria=local_criteria.vocab
            )

            if len(key_match) > 0:
                vardict[key] = key_match

        return vardict


def _get_axis_coord(obj: Union[DataFrame, Series], key: str) -> list:
    """
    Translate from axis or coord name to variable name. After matching based on coordinate_criteria,
    if there are no matches for key, then guess_regex is used to search for matches.

    Parameters
    ----------
    obj : DataArray, Dataset
        DataArray belonging to the coordinate to be checked
    key : str, ["X", "Y", "Z", "T", "longitude", "latitude", "vertical", "time"]
        key to check for.

    Returns
    -------
    List[str], Variable name(s) in parent xarray object that matches axis or coordinate `key`

    Notes
    -----
    This functions checks for the following attributes in order
    - `standard_name` (CF option)
    - `_CoordinateAxisType` (from THREDDS)
    - `axis` (CF option)
    - `positive` (CF standard for non-pressure vertical coordinate)

    References
    ----------
    MetPy's parse_cf
    """

    valid_keys = _COORD_NAMES + _AXIS_NAMES
    if key not in valid_keys:
        raise KeyError(
            f"cf_xarray did not understand key {key!r}. Expected one of {valid_keys!r}"
        )

    # search_in = set()
    # attrs_or_encoding = ChainMap(obj.attrs, obj.encoding)
    # coordinates = attrs_or_encoding.get("coordinates", None)

    # # Handles case where the coordinates attribute is None
    # # This is used to tell xarray to not write a coordinates attribute
    # if coordinates:
    #     search_in.update(coordinates.split(" "))
    # if not search_in:
    #     search_in = set(obj.coords)

    # # maybe only do this for key in _AXIS_NAMES?
    # search_in.update(obj.indexes)

    # search_in = search_in & set(obj.coords)
    results: set = set()
    for col in obj.columns:
        # var = obj.coords[coord]
        if key in coordinate_criteria:
            # import pdb; pdb.set_trace()
            for criterion, expected in coordinate_criteria[key].items():
                # allow for the column header having a space in it that separate
                # the name from the units, for example
                strings = col.split()
                for string in strings:
                    string = string.lower()
                    if string.startswith("(") and string.endswith(")"):
                        if string.strip(")(") in expected:
                            results.update((col,))
                    if string in expected:
                        # if col.attrs.get(criterion, None) in expected:
                        results.update((col,))
                    # if criterion == "units":
                    #     # deal with pint-backed objects
                    #     units = getattr(col.data, "units", None)
                    #     if units in expected:
                    #         results.update((col,))

        # also use the guess_regex approach by default, but only if no results so far
        # this takes the logic from cf-xarray guess_coord_axis
        if len(results) == 0:
            if obj[col].ndim == 1 and _is_datetime_like(obj[col]):
                results.update((col,))
                continue  # prevent second detection

            pattern = guess_regex[key]
            if pattern.match(col.lower()):
                results.update((col,))

    return list(results)


def _get_all(obj: DataFrame, key: str) -> List[str]:
    """
    One or more of ('X', 'Y', 'Z', 'T', 'longitude', 'latitude', 'vertical', 'time',
    'area', 'volume'), or arbitrary measures, or standard names
    """
    all_mappers = (
        _get_custom_criteria,
        # functools.partial(_get_custom_criteria, criteria=cf_role_criteria),
        _get_axis_coord,
        # _get_measure,
        # _get_with_standard_name,
    )
    results = apply_mapper(all_mappers, obj, key, error=False, default=None)
    return list(set(results))


def apply_mapper(
    mappers: Union[Mapper, Tuple[Mapper, ...]],
    obj: DataFrame,
    key: Hashable,
    error: bool = True,
    default: Any = None,
) -> List[Any]:
    """
    Applies a mapping function; does error handling / returning defaults.
    Expects the mapper function to raise an error if passed a bad key.
    It should return a list in all other cases including when there are no
    results for a good key.
    """

    if not isinstance(key, Hashable):
        if default is None:
            raise ValueError(
                "`default` must be provided when `key` is not not a valid DataArray name (of hashable type)."
            )
        return list(always_iterable(default))

    default = [] if default is None else list(always_iterable(default))

    def _apply_single_mapper(mapper):

        try:
            results = mapper(obj, key)
        except (KeyError, ValueError) as e:
            if error or "I expected only one." in repr(e):
                raise e
            else:
                results = []
        return results

    if not isinstance(mappers, Iterable):
        mappers = (mappers,)

    # apply a sequence of mappers
    # if the mapper fails, it *should* return an empty list
    # if the mapper raises an error, that is processed based on `error`
    results = []
    for mapper in mappers:
        results.append(_apply_single_mapper(mapper))

    flat = list(itertools.chain(*results))
    # # de-duplicate
    # if all(not isinstance(r, DataArray) for r in flat):
    #     results = list(set(flat))
    # else:
    #     results = flat
    results = flat

    nresults = any(bool(v) for v in [results])
    if not nresults:
        if error:
            raise KeyError(
                f"cf-xarray cannot interpret key {key!r}. Perhaps some needed attributes are missing."
            )
        else:
            # none of the mappers worked. Return the default
            return default
    return results


# Already use match_criteria_key in other functions, and it is a bit more generic so can be used
# without accessor.
def _get_custom_criteria(obj: DataFrame, key: str, criteria=None) -> List[str]:

    results = match_criteria_key(obj.columns, key, criteria, split=True)
    return results



"""
Utilities for cf-pandas.
"""

from collections import ChainMap
from typing import Any, Iterable, Optional, Union

import numpy as np
import pandas as pd
import regex
from pandas import Series

from .options import OPTIONS


def always_iterable(obj: Any, allowed=(tuple, list, set, dict)) -> Iterable:
    """This is from cf-xarray."""
    return [obj] if not isinstance(obj, allowed) else obj


def astype(value, type_):
    """Return `value` as type `type_`.
    Particularly made to work correctly for returning string, `PosixPath`, or `Timestamp` as list.
    """

    if not isinstance(value, type_):
        import pathlib

        if type_ == list and isinstance(value, (str, pathlib.PurePath, pd.Timestamp)):
            return [value]
        return type_(value)
    return value


def set_up_criteria(criteria: Union[dict, Iterable] = None) -> ChainMap:
    """Get custom criteria from options.

    Parameters
    ----------
    criteria : dict, optional
        Criteria to use to map from variable to attributes describing the variable. If user has defined
        custom_criteria, this will be used by default.

    Returns
    -------
    ChainMap
        Criteria
    """

    if criteria is None:
        if not OPTIONS["custom_criteria"]:
            raise ValueError(
                "criteria needs to be defined either using set_options or directly input."
            )
        criteria_it = OPTIONS["custom_criteria"]
    else:
        criteria_it = always_iterable(criteria, allowed=(tuple, list, set))

    # # Add in coordinate_criteria to be able to identify coordinates too
    # criteria_it[0].update(coordinate_criteria)

    return ChainMap(*criteria_it)


def match_criteria_key(
    available_values: list,
    keys_to_match: Union[str, list],
    criteria: Optional[dict] = None,
    split: bool = False,
) -> list:
    """Use criteria to choose match to key from available available_values.

    Parameters
    ----------
    available_values: list
        String or list of strings to compare against list of category values. They should be keys in `criteria`.
    keys_to_match : str, list
        Key(s) from criteria to match with available_values.
    criteria : dict, optional
        Criteria to use to map from variable to attributes describing the variable. If user has defined custom_criteria, this will be used by default.
    split : bool, optional
        If split is True, split the available_values by white space before performing matches. This is helpful e.g. when columns headers have the form "standard_name (units)" and you want to match standard_name.

    Returns
    -------
    list
        Values from available_values that match keys_to_match, according to criteria.

    Notes
    -----
    This uses logic from `cf-xarray`.
    """

    custom_criteria = set_up_criteria(criteria)

    keys_to_match = astype(keys_to_match, list)
    results = []
    for key in keys_to_match:

        if custom_criteria is not None and key in custom_criteria:
            # criterion is the attribute type — in this function we don't use it,
            # instead we use all the patterns available in criteria to match with available_values
            for criterion, patterns in custom_criteria[key].items():
                if split:
                    results.extend(
                        list(
                            set(
                                [
                                    value
                                    for value in available_values
                                    for value_part in value.split()
                                    if regex.match(patterns, value_part)
                                ]
                            )
                        )
                    )

                else:
                    results.extend(
                        list(
                            set(
                                [
                                    value
                                    for value in available_values
                                    if regex.match(patterns, value)
                                ]
                            )
                        )
                    )

        # catch scenario that user input valid reader variable names
        else:
            if key in available_values:
                results.append(key)
    return list(set(results))


def standard_names():
    """Returns list of CF standard_names.

    Returns
    -------
    list
        All CF standard_names
    """

    import requests
    from bs4 import BeautifulSoup

    url = "https://cfconventions.org/Data/cf-standard-names/79/src/cf-standard-name-table.xml"
    req = requests.get(url)
    soup = BeautifulSoup(req.content, features="xml")

    standard_names = [entry.get("id") for entry in soup.find_all("entry")]

    return standard_names


def _is_datetime_like(da: Series) -> bool:
    if np.issubdtype(da.dtype, np.datetime64) or np.issubdtype(
        da.dtype, np.timedelta64
    ):
        return True

    return False



"""
Criteria for identifying axes and coordinate variables.
Reused with modification from cf-xarray which modified from MetPy under the terms of the BSD 3-Clause License.
Copyright (c) 2017 MetPy Developers.
"""

import re
from typing import Mapping, MutableMapping, Tuple

coordinate_criteria: MutableMapping[str, MutableMapping[str, Tuple]] = {
    "latitude": {
        "standard_name": ("latitude",),
        "units": (
            "degree_north",
            "degree_N",
            "degreeN",
            "degrees_north",
            "degrees_N",
            "degreesN",
        ),
        "_CoordinateAxisType": ("Lat",),
    },
    "longitude": {
        "standard_name": ("longitude",),
        "units": (
            "degree_east",
            "degree_E",
            "degreeE",
            "degrees_east",
            "degrees_E",
            "degreesE",
        ),
        "_CoordinateAxisType": ("Lon",),
    },
    "Z": {
        "standard_name": (
            "model_level_number",
            "atmosphere_ln_pressure_coordinate",
            "atmosphere_sigma_coordinate",
            "atmosphere_hybrid_sigma_pressure_coordinate",
            "atmosphere_hybrid_height_coordinate",
            "atmosphere_sleve_coordinate",
            "ocean_sigma_coordinate",
            "ocean_s_coordinate",
            "ocean_s_coordinate_g1",
            "ocean_s_coordinate_g2",
            "ocean_sigma_z_coordinate",
            "ocean_double_sigma_coordinate",
        ),
        "_CoordinateAxisType": (
            "GeoZ",
            "Height",
            "Pressure",
        ),
        "axis": ("Z",),
        "cartesian_axis": ("Z",),
        "grads_dim": ("z",),
    },
    "vertical": {
        "standard_name": (
            "air_pressure",
            "height",
            "depth",
            "geopotential_height",
            # computed dimensional coordinate name
            "altitude",
            "height_above_geopotential_datum",
            "height_above_reference_ellipsoid",
            "height_above_mean_sea_level",
        ),
        "positive": ("up", "down"),
    },
    "X": {
        "standard_name": ("projection_x_coordinate",),
        "_CoordinateAxisType": ("GeoX",),
        "axis": ("X",),
        "cartesian_axis": ("X",),
        "grads_dim": ("x",),
    },
    "Y": {
        "standard_name": ("projection_y_coordinate",),
        "_CoordinateAxisType": ("GeoY",),
        "axis": ("Y",),
        "cartesian_axis": ("Y",),
        "grads_dim": ("y",),
    },
    "T": {
        "standard_name": ("time",),
        "_CoordinateAxisType": ("Time",),
        "axis": ("T",),
        "cartesian_axis": ("T",),
        "grads_dim": ("t",),
    },
}

coordinate_criteria["time"] = coordinate_criteria["T"]

# "long_name" and "standard_name" criteria are the same. For convenience.
for coord, attrs in coordinate_criteria.items():
    coordinate_criteria[coord]["long_name"] = coordinate_criteria[coord][
        "standard_name"
    ]
coordinate_criteria["X"]["long_name"] += ("cell index along first dimension",)
coordinate_criteria["Y"]["long_name"] += ("cell index along second dimension",)

guess_regex = {
    "time": re.compile("\\bt\\b|(time|min|hour|day|week|month|year)[0-9]*"),
    "Z": re.compile(
        "(z|nav_lev|gdep|lv_|[o]*lev|bottom_top|sigma|h(ei)?ght|altitude|depth|"
        "isobaric|pres|isotherm)[a-z_]*[0-9]*"
    ),
    "Y": re.compile("y|j|nlat|nj"),
    "latitude": re.compile("y?(nav_lat|lat|gphi)[a-z0-9]*"),
    "X": re.compile("x|i|nlon|ni"),
    "longitude": re.compile("x?(nav_lon|lon|glam)[a-z0-9]*"),
}
guess_regex["T"] = guess_regex["time"]



"""Class for creating and working with vocabularies."""

import json
import pathlib
from collections import defaultdict
from typing import DefaultDict, Dict, Optional, Sequence, Union

from .utils import astype


class Vocab(object):
    """Class to handle vocabularies."""

    def __init__(self, openname: Optional[str] = None):
        self.vocab: DefaultDict[str, Dict[str, str]]
        if openname is not None:
            self.vocab = defaultdict(dict, self.open_file(openname))
        else:
            self.vocab = defaultdict(dict)

    def __repr__(self):
        """Representation."""
        return dict(self.vocab).__repr__()

    def make_entry(
        self, nickname: str, expressions: Union[str, list], attr: str = "standard_name"
    ):
        """Make an entry for vocab.

        Parameters
        ----------
        nickname: str
            The nickname to call the variable being represented in this entry.
        expressions: str, list
            Regular expression(s) to use to select out the variable in a regex match. Multiple expressions input in a list are piped together to create one str of expressions.
        attr: str
            What attribute to identify the regular expressions with. Default is "standard_name", but other reasonable options are any variable attributes in a netcdf file such as "units", "name", and "long_name".

        Examples
        --------

        The following creates an entry in the vocabulary stored in `vocab.vocab`. It doesn't print the entry but it has been pasted in below the example to show what it looks like.

        >>> import cf_pandas as cfp
        >>> vocab = cfp.Vocab()
        >>> vocab.make_entry("temp", ["a","b"], attr="name")
        {'temp': {'standard_name': 'a|b'}})
        """

        expressions = astype(expressions, list)
        entry: DefaultDict[str, Dict[str, str]] = defaultdict(dict)
        entry[nickname][attr] = "|".join(expressions)
        self.__iadd__(entry)

        return self

    def add(
        self, other_vocab: Union[DefaultDict[str, Dict[str, str]], "Vocab"], method: str
    ) -> "Vocab":
        """Add two Vocab objects together...

        by adding their `.vocab`s together. Expressions are piped together but otherwise not changed.
        This is used for both `__add__` and `__iadd__`.

        Parameters
        ----------
        other_vocab: Vocab
            Other Vocab object to combine with.
        method : str
            Whether to run as "add" which returns a new Vocab object or "iadd" which adds to the original object.

        Returns
        -------
        Vocab
            vocab + other_vocab either as a new object or in place.
        """

        if isinstance(other_vocab, Vocab):
            other_vocab = other_vocab.vocab

        if method == "add":
            output = Vocab()
        elif method == "iadd":
            output = self

        nicknames = set(list(self.vocab.keys()) + list(other_vocab.keys()))
        for nickname in nicknames:

            # gather all attributes under nickname as a set to compare their expressions
            attributes = set(
                list(self.vocab[nickname].keys()) + list(other_vocab[nickname].keys())
            )

            # pipe together expressions for nickname-attribute pairs
            for attribute in attributes:
                new_expressions = (
                    self.vocab[nickname].get(attribute, "")
                    + "|"
                    + other_vocab[nickname].get(attribute, "")
                ).strip("|")
                output.vocab[nickname][attribute] = new_expressions
        return output

    def __add__(self, other_vocab: Union[DefaultDict[str, Dict[str, str]], "Vocab"]):
        """vocab1 + vocab2"""
        return self.add(other_vocab, "add")

    def __iadd__(self, other_vocab: Union[DefaultDict[str, Dict[str, str]], "Vocab"]):
        """vocab1 += vocab2"""
        return self.add(other_vocab, "iadd")

    def __radd__(
        self, other_vocab: Union[DefaultDict[str, Dict[str, str]], "Vocab"]
    ) -> "Vocab":
        """right add?"""
        return self.__add__(other_vocab)

    def save(self, savename: Union[str, pathlib.PurePath]):
        """Save to file.

        Parameters
        ----------
        savename: str, PurePath
            Filename to save to.
        """
        a_file = open(astype(savename, pathlib.PurePath).with_suffix(".json"), "w")
        json.dump(self.vocab, a_file)
        a_file.close()

    def open_file(self, openname: Union[str, pathlib.PurePath]):
        """Open previously-saved vocab.

        Parameters
        ----------
        openname: str
            Where to find vocab to open.
        """
        return json.loads(
            open(pathlib.PurePath(openname).with_suffix(".json"), "r").read()
        )


def merge(vocabs: Sequence[Vocab]) -> Vocab:
    """Add together multiple Vocab objects.

    Parameters
    ----------
    vocabs : Sequence[Vocab]
        Sequence of Vocab objects to merge.

    Returns
    -------
    Vocab
        Single Vocab object made up of input vocabs.
    """

    final_vocab = Vocab()
    for vocab in vocabs:
        final_vocab += vocab
    return final_vocab



"""
Copying options from cf-xarray options.py
"""

import copy
from typing import Any, MutableMapping

import cf_pandas as cfp

OPTIONS: MutableMapping[str, Any] = {
    "custom_criteria": [],
    # "warn_on_missing_variables": True,
}


class set_options:
    """Set options for cf-xarray in a controlled context.
    Parameters
    ----------
    custom_criteria : dict
        Translate from axis, coord, or custom name to
        variable name optionally using ``custom_criteria``. Default: [].
    warn_on_missing_variables : bool
        Whether to raise a warning when variables referred to in attributes
        are not present in the object.
    Examples
    --------
    You can use ``set_options`` either as a context manager:
    >>> import numpy as np
    >>> import xarray as xr
    >>> my_custom_criteria = {"ssh": {"name": "elev$"}}
    >>> ds = xr.Dataset({"elev": np.arange(1000)})
    >>> with cf_xarray.set_options(custom_criteria=my_custom_criteria):
    ...     xr.testing.assert_identical(ds["elev"], ds.cf["ssh"])
    ...
    Or to set global options:
    >>> cf_xarray.set_options(custom_criteria=my_custom_criteria)
    >>> xr.testing.assert_identical(ds["elev"], ds.cf["ssh"])
    """

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k not in OPTIONS:
                raise ValueError(
                    f"argument name {k!r} is not in the set of valid options {set(OPTIONS)!r}"
                )
            self.old[k] = OPTIONS[k]
        self._apply_update(kwargs)

    def _apply_update(self, options_dict):
        options_dict = copy.deepcopy(options_dict)
        for k, v in options_dict.items():
            if k == "custom_criteria":
                options_dict["custom_criteria"] = cfp.always_iterable(
                    options_dict["custom_criteria"], allowed=(tuple, list, set)
                )
        OPTIONS.update(options_dict)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        self._apply_update(self.old)



"""Test Reg class."""

import pandas as pd
import pytest
from pandas import testing as tm

import cf_pandas as cfp

strings = [
    "sea_water_temperature",
    "sea_water_temperature [celsius]",
    "water_temperature",
    "temperature_qc",
]
df = cfp.astype(strings, pd.Series)


def test_exclude():
    reg = cfp.Reg(exclude="qc")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = [
        "sea_water_temperature",
        "sea_water_temperature [celsius]",
        "water_temperature",
    ]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    # make sure this approach yields identical results, and use lists too
    reg2 = cfp.Reg()
    reg2.exclude(["qc"])
    dfmatch2 = df[df.str.match(reg2.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)

    # make sure inputting None and "" both yield the full list
    reg3 = cfp.Reg(exclude=None)
    dfmatch3 = df[df.str.match(reg3.pattern())]
    tm.assert_series_equal(dfmatch3, df, check_index=False)

    reg4 = cfp.Reg(exclude="")
    dfmatch4 = df[df.str.match(reg4.pattern())]
    tm.assert_series_equal(dfmatch4, df, check_index=False)


def test_exclude_two():
    reg = cfp.Reg(exclude=["qc", "sea"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_include_exact():
    reg = cfp.Reg(include_exact="water_temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg()
    reg.include_exact("water_temperature")
    dfmatch2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)


def test_include_exact_two():
    reg = cfp.Reg(include_exact="sea_water_temperature")
    with pytest.raises(ValueError):
        reg.include_exact("qc")


def test_include_exact_list():
    with pytest.raises(TypeError):
        reg = cfp.Reg(include_exact=["sea_water_temperature"])


def test_include():
    reg = cfp.Reg(include="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, df, check_index=False)

    reg = cfp.Reg()
    reg.include("temperature")
    dfmatch2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)

    reg = cfp.Reg(include_or="temperature")
    dfmatch_or = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch_or, check_index=False)

    reg = cfp.Reg()
    reg.include_or("temperature")
    dfmatch_or2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch_or2, check_index=False)


def test_include_two():
    reg = cfp.Reg(include=["water", "temp"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = [
        "sea_water_temperature",
        "sea_water_temperature [celsius]",
        "water_temperature",
    ]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_include_or_two():
    reg = cfp.Reg(include_or=["qc", "celsius"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]", "temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_include_end():
    reg = cfp.Reg(include_end="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature", "water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg()
    reg.include_end("temperature")
    dfmatch2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)


def test_include_end_two():
    reg = cfp.Reg(include_end="sea_water_temperature")
    with pytest.raises(ValueError):
        reg.include_end("qc")


def test_include_end_list():
    with pytest.raises(TypeError):
        reg = cfp.Reg(include_end=["sea_water_temperature"])


def test_exclude_include():
    reg = cfp.Reg(exclude="qc", include="sea_water_temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature", "sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(exclude="qc", include_or=["sea", "temp"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = [
        "sea_water_temperature",
        "sea_water_temperature [celsius]",
        "water_temperature",
    ]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(exclude="qc", include=["sea", "temp"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature", "sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_include_end():
    reg = cfp.Reg(exclude="sea", include_end="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_include_exact():
    with pytest.raises(ValueError):
        reg = cfp.Reg(exclude="sea", include_exact="temperature")


def test_include_include_exact():
    with pytest.raises(ValueError):
        reg = cfp.Reg(include="sea", include_exact="temperature")


def test_include_end_include_exact():
    with pytest.raises(ValueError):
        reg = cfp.Reg(include_end="sea", include_exact="temperature")


def test_include_include_end():
    reg = cfp.Reg(include="sea", include_end="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_include_start():
    reg = cfp.Reg(include_start="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg()
    reg.include_start("temperature")
    dfmatch2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)


def test_include_start_two():
    reg = cfp.Reg(include_start="sea_water_temperature")
    with pytest.raises(ValueError):
        reg.include_start("qc")


def test_include_start_list():
    with pytest.raises(TypeError):
        reg = cfp.Reg(include_start=["sea_water_temperature"])


def test_exclude_include_start():
    reg = cfp.Reg(exclude="celsius", include_start="sea")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_include_start_include_exact():
    with pytest.raises(ValueError):
        reg = cfp.Reg(include_start="sea", include_exact="temperature")


def test_include_include_start():
    reg = cfp.Reg(include="celsius", include_start="sea")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_start():
    reg = cfp.Reg(exclude_start="sea")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature", "temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg()
    reg.exclude_start("sea")
    dfmatch2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)


def test_exclude_start_two():
    reg = cfp.Reg(exclude_start=["water", "sea"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_start_include():
    reg = cfp.Reg(exclude_start="sea", include="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature", "temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(
        exclude_start=["water", "temperature"], include_or=["temp", "celsius"]
    )
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature", "sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(exclude_start=["water", "temperature"], include=["temp", "celsius"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_start_include_end():
    reg = cfp.Reg(exclude_start="sea", include_end="qc")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_start_include_exact():
    with pytest.raises(ValueError):
        reg = cfp.Reg(exclude_start="sea", include_exact="temperature")


def test_exclude_start_include_start():
    reg = cfp.Reg(exclude_start="sea", include_start="water")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_end():
    reg = cfp.Reg(exclude_end="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]", "temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg()
    reg.exclude_end("temperature")
    dfmatch2 = df[df.str.match(reg.pattern())]
    tm.assert_series_equal(dfmatch, dfmatch2, check_index=False)


def test_exclude_end_two():
    reg = cfp.Reg(exclude_end=["temperature", "qc"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(exclude_end=["temperature", r"\[celsius\]"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_end_include():
    reg = cfp.Reg(exclude_end="temperature", include="celsius")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(exclude_end="temperature", include_or=["temp", "celsius"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]", "temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(exclude_end="temperature", include=["temp", "celsius"])
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_end_include_end():
    reg = cfp.Reg(exclude_end="temperature", include_end="qc")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_end_include_exact():
    with pytest.raises(ValueError):
        reg = cfp.Reg(exclude_end="sea", include_exact="temperature")


def test_exclude_end_include_start():
    reg = cfp.Reg(exclude_end="temperature", include_start="temperature")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["temperature_qc"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_start_end():
    reg = cfp.Reg(
        exclude_end=r"\[celsius\]", exclude_start="temperature", exclude="sea"
    )
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_include_start_end():
    reg = cfp.Reg(include_or=["water", "celsius"], exclude="qc", include_start="sea")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature", "sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(include=["water", "celsius"], exclude="qc", include_start="sea")
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature [celsius]"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_exclude_start_end_include_start_end():
    reg = cfp.Reg(
        exclude_start="temperature",
        exclude="celsius",
        include="water",
        include_start="sea",
    )
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)


def test_ignore_case():
    strings = [
        "SEA_water_temperature",
        "sea_water_temperature",
    ]
    df = cfp.astype(strings, pd.Series)
    reg = cfp.Reg(
        include_start="sea",
        ignore_case=True,
    )
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["SEA_water_temperature", "sea_water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)

    reg = cfp.Reg(
        include_start="sea",
        ignore_case=False,
    )
    dfmatch = df[df.str.match(reg.pattern())]
    matches = ["sea_water_temperature"]
    tm.assert_series_equal(dfmatch, pd.Series(matches), check_index=False)
